#!perl -w
#
# Image -- what should go on the initrd image
#   Copyright (C) 2005  Erik van Konijnenburg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
use strict;
use warnings;
use File::Copy;
use Base;
package Image;
use base 'Obj';

sub fill {
	my $self = shift;
	$self->SUPER::fill();
	$self->{directories} = {};
	$self->{files} = {};
	$self->{devices} = {};
	$self->{symlinks} = {};
	$self->{scripts} = {};
}

#
# addFile -- mark listed file, it should be copied to image.
# If we add a symlink, make sure it stays a symlink in the target
# image, and that both link and target are included.  This matters
# eg for LVM, that has lots of symlinks to a single executable,
# with behaviour determined by the name of the link.
#
sub addFile {
	my ($self, $file) = @_;
	Base::assert (Base::isAbsolute ($file));

	$self->addParents ($file);
	if (-l $file) {
		if (! exists ($self->{symlinks}{$file})) {
			# careful: in order to break loops, we must
			# first mark the symlink, *then* follow it.
			# We cannot warn, since we do not know whether
			# the symlink was found in the course of symlink
			# resolution, or whether multiple executables
			# have the same symlink as a shared library.
			#
			# NOTE: Perhaps maintain an explicit stack of
			# symlinks being resolved in order to be able
			# to warn about loops?
			#
			my $link = readlink ($file);
			if (! defined ($link)) {
				Base::fatal ("could not follow symlink $file");
			}
			if ($link !~ m!^([\w./-]+)$!) {
				# taint check.
				Base::fatal ("Strange character in symlink target for $file");
			}
			$link = $1;

			Base::debug ("to image: $file (symlink)");
			$self->{symlinks}{$file} = $link;
			my $target = Base::expandLink ($file);
			$self->addFile ($target);
		}
	}
	elsif (-b $file) {
		my $devno = Base::devno ($file);
		Base::debug ("to image: $file (blockdev)");
		$self->{devices}{$file} = "b:$devno";
	}
	elsif (-c $file) {
		my $devno = Base::devno ($file);
		Base::debug ("to image: $file (chardev)");
		$self->{devices}{$file} = "c:$devno";
	}
	elsif (-d $file) {
		$self->addDirectory ($file);
	}
	elsif (-f $file) {
		Base::debug ("to image: $file");
		$self->{files}{$file}++;
		if (-x $file) {
			$self->addLibraries ($file);
		}
	}
	elsif (! -e $file) {
		Base::fatal ("missing file requested for image: $file");
	}
	else {
		Base::fatal ("odd file mode: $file");
	}
}

sub addDirectory {
	my ($self, $directory) = @_;
	Base::assert (Base::isAbsolute ($directory));

	$self->addParents ($directory);
	Base::debug ("to image: $directory (directory)");
	$self->{directories}{$directory}++;
}

# addParents -- given a pathname to be included on
# initrd image, ensure that parent directories are also created.
sub addParents {
	my ($self, $path) = @_;
	Base::assert (Base::isAbsolute ($path));

	my @components = split (/\/+/, $path);
	for my $idx (1 .. ($#components -1)) {
		my $dir = join ('/', @components[0 .. $idx]);
		$self->{directories}{$dir}++;
	}
}

# addLibraries -- given an executable to be included on
# initrd image, ensure that shared libraries are also included.
# If the executable is a shell script or statically linked,
# nothing needs to be done.
sub addLibraries {
	my ($self, $executable) = @_;
	my $in;

	if ($executable !~ m!^([.\w/-]+)$!) {
		# taint check.
		Base::fatal ("odd characters in executable name: $executable");
	}
	$executable = $1;

	if (! open ($in, "-|", "/usr/bin/ldd $executable")) {
		Base::fatal ("can't read ldd for $executable");
	}
	while (defined (my $line = <$in>)) {
		chomp $line;
		last if ($line =~ /statically linked/);
		if ($line =~ /not a dynamic executable/) {
			#
			# Happens when doing ldd on a shell script
			# or kernel module (these have x-bit on in FC3,
			# no idea why they would do that).  Ldd exit status
			# will be non-zero; return avoids the test for that.
			#
			return;
		}
		elsif ($line =~ m!.* => (/.+) \(0x[0-9a-fA-F]+\)$!) {
			#
			# in ldd 2.3.2 (Debian) output looks like this:
			# /lib/ld-linux.so.2 => /lib/ld-linux.so.2 (0xb7fea000)
			#
			$self->addFile ($1);
		}
		elsif ($line =~ m!^\s*(/[^\s]+) \(0x[0-9a-fA-F]+\)$!) {
			#
			# But ldd 2.3.3 (FC3) has this:
			# /lib/ld-linux.so.2 (0x0056f000)
			#
			$self->addFile ($1);
		}
		elsif ($line =~ m!.* =>  \(0x[0-9a-fA-F]+\)$!) {
			#
			# And eg amd64 also has this:
			#	linux-gate.so.1 =>  (0x00000000)
			#
			# note the double space.  This is a section not
			# loaded from a file but mapped in by the kernel.
			# The kernel puts some code with proper syscall
			# linkage here.
			# See http://lkml.org/lkml/2003/6/18/156
			#
			next;
		}
		else {
			Base::fatal ("unexpected ldd output for $executable: $line");
		}
	}
	if (! close ($in)) {
		Base::fatal ("could not read ldd for $executable");
	}
}

#
# addScriptLine -- given scriptname and text, append text to the
# text to be included in script.
#
sub addScriptLine {
	my ($self, $scriptname, $text) = @_;
	Base::assert (Base::isAbsolute ($scriptname));

	$self->addParents ($scriptname);
	$self->{scripts}{$scriptname} .= "$text";
}


#
# buildHere -- given pathname of a destination directory,
# create that directory, with all files and scripts prescribed
# by self in it.
# The top directory is expected to be created by File::Temp::tempfile.
#
sub buildHere {
	my ($self, $dest) = @_;

	if (! -d $dest) {
		Base::fatal ("Build directory not found: $dest");
	}
	if (! chmod (0755, $dest)) {
		Base::fatal ("Could not chmod $dest");
	}

	for my $d (sort keys %{$self->{directories}}) {
		if (! mkdir ("$dest$d")) {
			Base::fatal ("Could not mkdir $dest$d");
		}
		if (! chmod (0755, "$dest$d")) {
			Base::fatal ("Could not chmod $dest$d");
		}
	}

	for my $f (sort keys %{$self->{files}}) {
		if (! File::Copy::copy ($f, "$dest$f")) {
			Base::fatal ("Could not copy $f to $dest$f");
		}
		if (-x $f) {
			if (! chmod (0555, "$dest$f")) {
			    Base::fatal ("Could not chmod 0555 $dest$f");
			}
		}
		else {
			if (! chmod (0444, "$dest$f")) {
			    Base::fatal ("Could not chmod 0444 $dest$f");
			}
		}
	}

	for my $d (sort keys %{$self->{devices}}) {
		my $typDevno = $self->{devices}{$d};
		if ($typDevno =~ /^(b|c):(\d+):(\d+)$/) {
			#
			# permission 0600 is restrictive but sufficient:
			# the devices created here are used only by root
			# on a temporary root, so we can get away with
			# /dev/null mode 0600.  More permissive modes
			# risk ending up with a /dev/hda1 0666 somewhere
			# in a temporary directory.
			#
			if (system ("/bin/mknod -m 0600 '$dest$d' $1 $2 $3") != 0) {
				Base::fatal ("Could not mknod $dest$d $1 $2 $3");
			}
		}
		else {
			Base::bug ("bad devno $typDevno for $d");
		}
	}

	for my $l (sort keys %{$self->{symlinks}}) {
		my $target = $self->{symlinks}{$l};
		if (! symlink ($target, "$dest/$l")) {
			Base::fatal ("Could symlink $target $dest$l");
		}
	}

	for my $scriptname (sort keys %{$self->{scripts}}) {
		my $script = $self->{scripts}{$scriptname};
		if (!open (OUT, ">", "$dest$scriptname")) {
			Base::fatal ("Can't create $dest$scriptname");
		}
		print OUT "$script";
		if (!close (OUT)) {
			Base::fatal ("Could not write $dest$scriptname");
		}
		if (! chmod (0555, "$dest$scriptname")) {
			Base::fatal ("Couldn't chmod 0555 $dest$scriptname");
		}
	}
}

1;
