#!perl -w
#
# ModDepTab -- encapsulate /lib/modules/2.6.x/modules.dep
#   Copyright (C) 2005  Erik van Konijnenburg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Most important routine is resolve(), which returns a list of
# modules with dependencies resolved.
#
# We remember underscore as hyphen.
#
# For every module, we have a list of prerequisite modules;
# these prerequisites are in no particular order: the prerequisites
# may have dependencies among themselves that are not satisfied
# by the order in the list.
#

use strict;
use warnings;
use Base;
use Conf;
use ModDir;
use KConfig;

package ModDepTab;

my $modDepTab = undef;

sub init () {
	if (defined ($modDepTab)) {
		return;
	}
	$modDepTab = {};
	my $name = Conf::get('modDep');
	my $modExt = Conf::get('modExt');
	if (! open (IN, "<", "$name")) {
		Base::fatal ("can't open $name");
	}
	while (defined (my $line = <IN>)) {
		chomp $line;
		$line =~ s/#.*//;
		$line =~ s/^\s+//;
		$line =~ s/\s+$//;
		next if $line eq "";
		if ($line =~ /^(.*):\s*(.*)$/) {
			my $key = $1;
			my $val = $2;
			my @fields = split (/\s+/, $val);
			$key =~ s!.*/!!;
			$key =~ s!\.$modExt$!!;
			$key =~ s!_!-!g;
			for my $i (0 .. $#fields) {
				$fields[$i] =~ s!.*/!!;
				$fields[$i] =~ s!\.$modExt$!!;
				$fields[$i] =~ s!_!-!g;
			}
			# print "$key: " . join (', ', @fields) . "\n";
			if (! defined (ModDir::findByModnam($key))) {
				Base::fatal ("bogus module $key in $name");
			}
			$modDepTab->{$key} = [ @fields ];
		}
		else {
			Base::fatal ("garbage in $name");
		}
		
	}
	if (! close (IN)) {
		Base::fatal ("could not read $name");
	}
}

sub all	() {
	init;
	return $modDepTab;
}

sub findByModnam ($) {
	my ($module) = @_;
	init;
	$module =~ s!_!-!g;
	my $result = $modDepTab->{$module};
	if (! defined ($result)) {
		Base::fatal ("can't find dependencies for $module");
	}
	return $result;
}


# return a copy of the list with prerequisite modules inserted
sub resolve ($) {
	my ($modList) = @_;
	my $result = [];
	resolve2 ($modList, $result);
	return $result;
}

#
# resolve2 -- given a list of modules, that may have dependencies
# (even among themselves) and a list of modules that is in proper order
# to be loaded without unresolved dependencies,
# add everything from the first list (plus their dependencies)
# to the second list, without introducing duplicates.
#
sub resolve2 ($$);
sub resolve2 ($$) {
	my ($modList, $done) = @_;
OUTER:	for my $module (@{$modList}) {
		#
		# Stuff that is compiled into the kernel cannot
		# add or resolve dependencies, and obviously don't
		# need to be loaded.  Drop them from the list.
		#
		if (KConfig::isBuiltIn ($module)) {
			next OUTER;
		}
		#
		# if the module is already there, don't add again
		#
		for my $d (@{$done}) {
			if ($d eq $module) {
				next OUTER;
			}
		}
		#
		# recursively add prerequisites
		#
		my $deps = ModDepTab::findByModnam($module);
		if (! defined ($deps)) {
			Base::fatal ("can't find dependencies for $module");
		}
		resolve2 ($deps, $done);
		#
		# and module itself
		#
		push @{$done}, $module;
	}
}

1;
