#!/usr/bin/perl
#
# Plan -- plan what to put on the initrd image
#   Copyright (C) 2005  Erik van Konijnenburg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
use strict;
use warnings;
use Base;
use FsTab;
use ActiveBlockDevTab;
use LvmTab;
use Hardware;
use ModDepTab;
use RaidTab;
use Image;
use Blacklist;
use ActionList;

package Plan;



#
# modulesExpand -- given a list of modules, return an ordered
# list of modules, with blacklisted items deleted,
# compiled in modules deleted, and any dependencies resolved.
# A module found by eg PCI matching and dropped because of blacklisting
# may be added later to resolve a dependency.
#
sub modulesExpand ($) {
	my ($modlist) = (@_);

	my $ml = [];
	for my $moduleName (@{$modlist}) {
		if (Blacklist::isBlacklisted ($moduleName)) {
			next;
		}
		push @{$ml}, $moduleName;
	}

	# resolve drops compiled-in stuff.
	my $fullModList = ModDepTab::resolve ($ml);

	return $fullModList;
}



#
# addModules -- given an actionList and list of modules,
# add actions to load them all, unless blacklisted or compiled in.
#
sub addModules ($$) {
	my ($actionList, $modList) = @_;

	my $fullModList = modulesExpand ($modList);
	for my $moduleName (@{$fullModList}) {
		my $path = ModDir::findByModnam ($moduleName);
		if (! defined ($path)) {
			Base::fatal ("Could not find module $moduleName");
		}
		$actionList->add ("insmod", $path);
	}
}


#
# addHardware -- given hardware path, potentially undef,
# add required actions to the action list.
#
sub addHardware ($$) {
	my ($actionList, $hardware) = @_;
	if (defined ($hardware)) {
		my $modList = Hardware::moduleList ($hardware);
		addModules ($actionList, $modList);
		Base::info ("hardware: completed $hardware");
	}
}


#
# addDevicePlan -- given an active block device, add to actionList
# a sequence to make it into a working device file, suitable for mounting.
# For eg LVM modules this means loading dm-mod, but also it
# means recursively making the underlying devices available.
# There is a possibility of loops here (consider an fstab
# where /a is a loopback mount of /a/disk.img); we want do
# detect these loops and produce an error.  Also, we want to
# avoid double work: if five lvm logical volumes use the same
# underlying raid device, there's not much point in running
# mdadm five times.
# There are some contraints on the name of the device file;
# see yspecial in ActiveBlockDev.
#
# actionList - add actions here
# device - the active block device to be made available.
# done - list of devices that can be assumed to be available;
#		updated in place.
# working - list of devices we're working on, excluding path.
#
sub addDevicePlan ($$$$);
sub addDevicePlan ($$$$) {
	my ($actionList, $device, $done, $working) = @_;

	for my $w (@{$working}) {
		if ($w eq $device) {
			Base::fatal ("loop detected: $device\n");
		}
	}
	for my $w (@{$done}) {
		if ($w eq $device) {
			return;
		}
	}

	my $name = $device->name;
	my $parent = $device->parent;
	my $devno = $device->devno;
	my $hardware = $device->hw;
	my $yspecial = $device->yspecial;

	if (defined ($parent)) {
		#
		# It's a partition.  To make it available, do the whole
		# device, then make a special file for the partition.
		#
		addDevicePlan($actionList,$parent,$done,[$device,@{$working}]);
		my $pname = $parent->name;
		$actionList->add ("mkbdev", $yspecial,
				sysname => "$pname/$name");
	}
	elsif ($name =~ /^dm-\d+$/) {
		#
		# To start an LVM device, make available the underlying
		# physical volumes, then start the volume group.
		# creating the block special file is a side effect of
		# starting the volume group.
		# The activeBlockDev->yspecial() method is choosen such
		# that the name returned matches the name of the block special
		# file generated by vgchange.
		# Physical volumes are named by the block
		# special file that gives access to it.
		#
		my $lv = LvmTab::findLVByDevno ($devno);
		if (! defined ($lv)) {
			Base::fatal ("Can't find LVM info for $name");
		}
		my $vgnam = $lv->vgnam;
		for my $physVol (@{LvmTab::findPVsByVgnam($vgnam)}) 
		{
			my $pdev = ActiveBlockDevTab::findByPath
							($physVol->pvnam);
			addDevicePlan ($actionList, $pdev, $done,
					[$device,@{$working}]);
		}
		addModules ($actionList, [ "dm-mod" ]);
		$actionList->add ("vgchange", $vgnam);
	}
	elsif ($name =~ /^md\d+$/) {
		#
		# To start an md raid device, start the underlying hardware,
		# load raid module, then do mdadm --assemble.
		#
		my $rd = RaidTab::findByDevno ($devno);
		if (! defined ($rd)) {
			Base::fatal ("Can't find Raid info for $name");
		}
		my $components = [];
		for my $subDiskPath (@{$rd->devices()}) {
			my $subDisk = ActiveBlockDevTab::findByPath
					($subDiskPath);
			addDevicePlan ($actionList, $subDisk,
					$done,[$device,@{$working}]);
			my $subName = $subDisk->yspecial;
			push @{$components}, {dev => $subName};
		}
		addModules ($actionList, [ $rd->level ]);

		my $uuid = $rd->uuid;
		my ($major, $minor) = ($devno =~ /(\d+):(\d+)/);

		$actionList->add ("mdadm", $yspecial,
			major => $major,
			minor => $minor,
			uuid => $uuid,
			components => $components,
			);
	}
	elsif ($name =~ /^hd[a-z]$/
		|| $name =~ /^sd[a-z]$/)
	{
		# IDE or SCSI.
		addHardware ($actionList, $hardware);
		$actionList->add ("mkbdev", $yspecial, sysname => "$name");
	}
	elsif ($name =~ /^fd\d+$/) {
		addModules ($actionList, [ "floppy" ]);
		$actionList->add ("mkbdev", $yspecial, sysname => "$name");
	}
	else {
		Base::fatal ("unsupported device required: $name");
	}

	push @{$done}, $device;
}


#
# addInputPlan -- list of actions to activate console,
# for now only keyboard, no framebuffer.
#
sub addInputPlan ($) {
	my ($actionList) = @_;
	for my $input (@{InputTab::all()}) {
		if (! $input->isKbd()) {
			next;
		}
		addHardware ($actionList, $input->hw);
	}
}


#
# makePlan -- return a list of actions to mount root.
#
sub makePlan () {
	my $actionList = ActionList->new();

	$actionList->add ("prologue", "");

	addInputPlan($actionList);

	my $root = FsTab::findByMountPoint('/');
	if (! defined ($root)) {
		fatal ("no root in fstab");
	}

	# swap space is needed for soft suspend,
	# computed like so, but unused for now.
	my @swapEntries = ();
	for my $entry (@{FsTab->all}) {
		if (! $entry->isRequiredAtBoot()) {
			next;
		}
		if ($entry->type eq "swap") {
			push @swapEntries, $entry;
		}
	}

	my $devicesDone = [];
	# for swsusp, use ($root, @swapEntries)
	for my $fsTabEntry ($root) {
		my $path = $fsTabEntry->blockDevPath();
		if (! defined ($path)) {
			my $str = $fsTabEntry->string;
			Base::fatal ("unsupported block device specification in $str");
		}
		my $abd = ActiveBlockDevTab::findByPath($path);
		if (! defined ($abd)) {
			my $str = $fsTabEntry->string;
			Base::fatal ("block device $path for $str unavailable");
		}
		addDevicePlan ($actionList, $abd, $devicesDone, []);
	}

	my $fsType = $root->type;
	addModules ($actionList, [ $fsType ]);

	my $path = $root->blockDevPath();
	my $yspecial = ActiveBlockDevTab::findByPath($path)->yspecial();
	my $opts = $root->opts->cmdLineVersion();

	# mount -n: don't write /etc/mtab (on ro filesys)
	$actionList->add ("mount", "/mnt",
		options => "-n -r -t $fsType $opts",
		device => $yspecial);

	$actionList->add ("postlude", "");

	return $actionList;
}


1;
