#!/usr/bin/perl
#
# Template -- for translation general intentions to exact initrd layout
#   Copyright (C) 2005  Erik van Konijnenburg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# These templates define the translation from general intentions in the
# plan to exact lists of files and script fragments to be included 
# on the initrd image.  Templates are indexed by an action name, and include:
# 	- files to be copied from current system to initrd;
# 	  often executables.
# 	- directories to be created
# 	- fragments to be appended to scripts on the image,
# 	  these are in a hash from script name to script content.
#
# Everything is processed by HTML::Template before being added to the
# image: script fragments, but also file names and script names.
# Leading spaces are dropped from template lines.
#
# If you want to implement busybox support, this is the place to do it.
#
# Obviously, the template names and the attributes used in the template
# should match what is provided by the planner.
#
#

use strict;
package Template;


my $templates = {
	prologue => {
		# In principle, we don't want device files on the initrd
		# image: device numbers can change between kernels.
		# Instead, create device files at boot time based on info
		# in sysfs.  However, /dev/null and /dev/console are needed
		# to mount sysfs.
		files => [
			'/dev/null',
			'/dev/console',
			'/bin/dash',
			'/bin/cat',
			'/bin/mkdir',
			'/bin/mount',
			'/bin/umount',
			'/bin/mknod',
			'/bin/sleep',
			'/usr/sbin/chroot',
			'/sbin/pivot_root',
			],

		directories => [
			'/mnt', '/proc', '/sys', '/dev', '/etc', '/var',
			],

		scripts => {
			'/linuxrc' => <<'//',
				#!/bin/dash
				set -x
				export PATH=/sbin:/bin
				mount -nt proc proc /proc
				root=$(cat proc/sys/kernel/real-root-dev)
				echo 256 > proc/sys/kernel/real-root-dev
				echo "linuxrc: root is $root"
				umount -n /proc
				exit 0
//

			'/sbin/init' => <<'//',
				#!/bin/dash
				set -x
				mkcdev () {
					file="$1"
					devnam="$2"
					devpair=`/bin/cat "/sys/class/$devnam/dev"`
					maj=${devpair%:*}
					min=${devpair#*:}
					/bin/mknod "$file" c $maj $min
				}
				mkbdev () {
					file="$1"
					devnam="$2"
					devpair=`/bin/cat "/sys/block/$devnam/dev"`
					maj=${devpair%:*}
					min=${devpair#*:}
					/bin/mknod "$file" b $maj $min
				}
				mount -nt sysfs sysfs /sys
				mount -nt proc proc /proc
				mount -nt tmpfs tmpfs /dev
//
		},
	},

	insmod => {
		files => [
			'<TMPL_VAR NAME=target>',
			'/sbin/insmod',
			],
		directories => [],
		scripts => {
			'/sbin/init' => <<'//',
				/sbin/insmod '<TMPL_VAR NAME=target>'
//
		},

	},

	mkbdev => {
		files => [],
		directories => [],
		scripts => {
			'/sbin/init' => <<'//',
				mkbdev '<TMPL_VAR NAME=target>' '<TMPL_VAR NAME=sysname>'
//
		},
	},

	vgchange => {
		files => [
			'/lib/lvm-200/vgchange',
			'/sbin/vgchange',
			],
		directories => [],
		scripts => {
			# We cannot make mapper/control any earlier,
			# since dm-mod needs to be loaded to be able
			# to determine devno.
			'/sbin/init' => <<'//',
				if [ ! -c /dev/mapper/control ]
				then
					/bin/mkdir /dev/mapper
					mkcdev /dev/mapper/control misc/device-mapper
				fi
				/bin/mount -nt tmpfs tmpfs /etc
				/bin/mount -nt tmpfs tmpfs /var
				/sbin/vgchange -a y '<TMPL_VAR NAME=target>'
				/bin/umount -n /var
				/bin/umount -n /etc
//
		},
	},

	mdadm => {
		files => [
			'/sbin/mdadm',
			],
		directories => [],
		scripts => {
			#
			# NOTE: mdadm can operate without knowledge
			# of device numbers; it will assign an unused
			# one to the new device.  Thus you could have
			# device name /dev/md/boot that is stable,
			# regardless of which other raid devices are
			# operational.  However, debian installer
			# consistently uses device names like /dev/md0,
			# where the name depends on earlier minors
			# that happen to have been assigned.
			# To work around this problem, we forego
			# the --auto option, and mknod the device
			# before putting an md under it.  Better hope
			# the major number is not going to change ...
			#
			'/sbin/init' => <<'//',
				mknod <TMPL_VAR NAME=target> b <TMPL_VAR NAME=major> <TMPL_VAR NAME=minor>
				mdadm --assemble <TMPL_VAR NAME=target> --uuid <TMPL_VAR NAME=uuid> \
					<TMPL_LOOP NAME=devices> <TMPL_VAR NAME=dev></TMPL_LOOP>
//
		}
	},

	mount => {
		files => [],
		directories => [],
		scripts => {
			'/sbin/init' => <<'//',
				mount <TMPL_VAR NAME=options> <TMPL_VAR NAME=device> <TMPL_VAR NAME=target>
//
		}
	},

	postlude => {
		files => [],
		directories => [],
		scripts => {
			'/sbin/init' => <<'//',
				echo "Debugging opportunity, type ^D to continue."
				/bin/dash
				echo "Starting pivot_root ..."
				cd /mnt
				# unmount /dev/twice, may have devfs planted by kernel
				/bin/umount -n /dev
				/bin/umount -n /dev
				/bin/umount -n /sys
				/bin/umount -n /proc
				/sbin/pivot_root . initrd
				exec initrd/usr/sbin/chroot . /sbin/init < dev/console > dev/console 2>&1
//
		}
	},
};

sub template ($) { return $templates->{$_[0]}; }


1;
