#!/usr/bin/perl
#
# Template -- for translation general intentions to exact initrd layout
#   Copyright (C) 2005  Erik van Konijnenburg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# These templates define the translation from general intentions in the
# plan to exact lists of files and script fragments to be included 
# on the initrd image.  Templates are indexed by an action name, and include:
# 	- files to be copied from current system to initrd;
# 	  often executables.
# 	- directories to be created
# 	- fragments to be appended to scripts on the image,
# 	  these are in a hash from script name to script content.
#
# Everything is processed by HTML::Template before being added to the
# image: script fragments, but also file names and script names.
# Leading spaces are dropped from template lines.
#
# If you want to implement busybox support, this is the place to do it.
#
# Obviously, the template names and the attributes used in the template
# should match what is provided by the planner.
#
#

use strict;
package Template;


my $templates = {
	prologue => {
		# In principle, we don't want device files on the initrd
		# image: device numbers can change between kernels.
		# Instead, create device files at boot time based on info
		# in sysfs.  However, /dev/null and /dev/console are needed
		# to mount sysfs.
		files => [
			'/dev/null',
			'/dev/console',
			'/bin/ash',
			'/bin/cat',
			'/bin/mkdir',
			'/bin/mount',
			'/bin/umount',
			'/bin/mknod',
			'/bin/sleep',
			],

		directories => [
			'/mnt', '/proc', '/sys', '/dev', '/etc', '/var',
			],

		scripts => {
			'/init' => <<'//',
				#!/bin/ash
				set -x
				mkcdev () {
					file="$1"
					devnam="$2"
					devpair=`/bin/cat "/sys/class/$devnam/dev"`
					maj=${devpair%:*}
					min=${devpair#*:}
					/bin/mknod "$file" c $maj $min
				}
				mkbdev () {
					file="$1"
					devnam="$2"
					devpair=`/bin/cat "/sys/block/$devnam/dev"`
					maj=${devpair%:*}
					min=${devpair#*:}
					/bin/mknod "$file" b $maj $min
				}
				mount -nt sysfs sysfs /sys
				mount -nt proc proc /proc

				# In Fedora FC3, this *must* have the name
				# 'none', otherwise start_udev, called from
				# rc.sysinit will mount a new tmpfs over it,
				# and you loose access to nodes such as
				#/dev/vg0/root, created by LVM in initrd.
				mount -nt tmpfs none /dev
//
		},
	},

	insmod => {
		files => [
			'<TMPL_VAR NAME=target>',
			'/sbin/insmod',
			],
		directories => [],
		scripts => {
			'/init' => <<'//',
				/sbin/insmod '<TMPL_VAR NAME=target>'
//
		},

	},

	mkbdev => {
		files => [],
		directories => [],
		scripts => {
			'/init' => <<'//',
				mkbdev '<TMPL_VAR NAME=target>' '<TMPL_VAR NAME=sysname>'
//
		},
	},

	vgchange => {
		#
		# Use static version.  Note that fedora starts with
		# empty /dev, so we must make device nodes explicitly.
		#
		files => [
			'/sbin/lvm',
			],
		directories => [],
		scripts => {
			# We cannot make mapper/control any earlier,
			# since dm-mod needs to be loaded to be able
			# to determine devno.
			'/init' => <<'//',
				if [ ! -c /dev/mapper/control ]
				then
					/bin/mkdir /dev/mapper
					mkcdev /dev/mapper/control misc/device-mapper
				fi
				/bin/mount -nt tmpfs tmpfs /etc
				/bin/mount -nt tmpfs tmpfs /var
				/sbin/lvm vgchange -a y '<TMPL_VAR NAME=target>'
				/sbin/lvm vgmknodes
				/bin/umount -n /var
				/bin/umount -n /etc
//
		},
	},

	mdadm => {
		files => [
			'/sbin/mdadm',
			],
		directories => [],
		scripts => {
			#
			# NOTE: mdadm can operate without knowledge
			# of device numbers; it will assign an unused
			# one to the new device.  Thus you could have
			# device name /dev/md/boot that is stable,
			# regardless of which other raid devices are
			# operational.  However, debian installer
			# consistently uses device names like /dev/md0,
			# where the name depends on earlier minors
			# that happen to have been assigned.
			# To work around this problem, we forego
			# the --auto option, and mknod the device
			# before putting an md under it.  Better hope
			# the major number is not going to change ...
			#
			'/init' => <<'//',
				mknod <TMPL_VAR NAME=target> b <TMPL_VAR NAME=major> <TMPL_VAR NAME=minor>
				mdadm --assemble <TMPL_VAR NAME=target> --uuid <TMPL_VAR NAME=uuid> \
					<TMPL_LOOP NAME=components> <TMPL_VAR NAME=dev></TMPL_LOOP>
//
		}
	},

	mount => {
		files => [],
		directories => [],
		scripts => {
			'/init' => <<'//',
				mount <TMPL_VAR NAME=options> <TMPL_VAR NAME=device> <TMPL_VAR NAME=target>
//
		}
	},

	postlude => {
		files => [
			'/bin/cp',
			'/usr/sbin/chroot',
			],
		directories => [],
		scripts => {
			#
			# fedora does not come with populated /dev 
			# on root and needs at least /dev/console.
			# Pass them via tmpfs.
			#
			# Do *not* do a pivot_root here.  We're
			# working from rootfs; a pivotroot would
			# make that end up in /initrd, which gets
			# unmounted by rc.sysinit, and without rootfs,
			# too much is broken to even do a panic.
			#
			# Instead mount-move the new root over the
			# old rootfs root, and chroot and cd into it.
			# All of these are needed to ensure all traces
			# of rootfs are gone before we enter late userspace.
			#
			# Rootfs should be cleaned up before moving
			# into late userspace, but I'm lazy.
			# We definitely need a small bit of C code
			# here to do the transition cleanly.
			# See nash, included with FC3 mkinitrd
			# or run-init, included with klibc for more
			# mature code.  (not to be confused with the
			# PAM related run_init.)
			# 
			# 

			'/init' => <<'//',
				echo "Debugging opportunity, ^D to continue."
				/bin/ash

				echo "Switching root ..."
				cd /mnt
				mkcdev /dev/null mem/null
				mkcdev /dev/zero mem/zero
				mkcdev /dev/console tty/console
				mkdir /dev/pts
				mkdir /dev/shm
				mount -n -t tmpfs --bind /dev /mnt/dev
				/bin/umount -n /sys
				/bin/umount -n /proc
				# NOTE: we need to make this part a C program.
				# if only because of this:
				/bin/cp /usr/sbin/chroot /mnt/dev/.
				mount -n --move . /
				# the cd / must come after the chroot.
				exec /dev/chroot . /bin/ash -c 'cd / && exec /sbin/init' < dev/console > dev/console 2>&1
//
		}
	},
};

sub template ($) { return $templates->{$_[0]}; }


1;
